/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include<stdexcept>
#include<iostream>
#include<utility>
#include"mobility.h"
#include"../common/prebb.h"
using std::cout;
using std::endl;
using std::move;
using std::logic_error;


Mobility::Mobility(const ChessPosition& x): cp(x), m_atta(x.attacks())
{
}

void Mobility::compute_white_pawn_attacks()
{
  BitBoard bb{cp.white_pawn()};
  m_white_pawn_attacks=bb<<7 & not_h_file;  
  m_white_pawn_attacks|=bb<<9 & not_a_file;  
  m_white_attacks=m_white_pawn_attacks;
}

void Mobility::compute_black_pawn_attacks()
{
  BitBoard bb{cp.black_pawn()};
  m_black_pawn_attacks=bb>>9 & not_h_file;  
  m_black_pawn_attacks|=bb>>7 & not_a_file;  
  m_black_attacks=m_black_pawn_attacks;
}

void Mobility::compute_white_knight_attacks()
{
  m_white_knight_attacks.clear();
  BitBoard bb{cp.white_knight()};
  while(bb){
    BitBoard s=bb.get_reset_square();
    BitBoard ats=m_atta.knight(s);
    m_white_knight_attacks|=ats;
    m_white_attacks|=ats;
    ats&=~m_black_pawn_attacks;
    ats&=~cp.white_pawn();
    m_safe_mobility+=knight_mobility[ats.popcount()];
  }
}

void Mobility::compute_black_knight_attacks()
{
  m_black_knight_attacks.clear();
  BitBoard s,ats;
  BitBoard bb{cp.black_knight()};
  while(bb){
    s=bb.get_reset_square();
    ats=m_atta.knight(s);
    m_black_knight_attacks|=ats;
    m_black_attacks|=ats;
    ats&=~m_white_pawn_attacks;
    ats&=~cp.black_pawn();
    m_safe_mobility-=knight_mobility[ats.popcount()];
  }
}

void Mobility::compute_white_bishop_attacks()
{
  m_white_bishop_attacks.clear();
  BitBoard bb{cp.white_bishop()};
  while(bb){
    BitBoard s=bb.get_reset_square();
    BitBoard ats=m_atta.bishop(s,cp.pieces() & ~cp.white_queen());
    m_white_bishop_attacks|=ats;
    m_white_attacks|=ats;
    ats&=~m_black_pawn_attacks;
    ats&=~cp.white_pawn();
    m_safe_mobility+=bishop_mobility[ats.popcount()];
  }
}

void Mobility::compute_black_bishop_attacks()
{
  m_black_bishop_attacks.clear();
  BitBoard bb{cp.black_bishop()};
  while(bb){
    BitBoard s=bb.get_reset_square();
    BitBoard ats=m_atta.bishop(s,cp.pieces() & ~cp.black_queen());
    m_black_bishop_attacks|=ats;
    m_black_attacks|=ats;
    ats&=~m_white_pawn_attacks;
    ats&=~cp.black_pawn();
    m_safe_mobility-=bishop_mobility[ats.popcount()];
  }
}

void Mobility::compute_white_rook_attacks()
{
  m_white_rook_attacks.clear();
  BitBoard bb{cp.white_rook()};
  while(bb){
    BitBoard s=bb.get_reset_square();
    BitBoard ats=m_atta.rook(s,cp.pieces() & ~cp.white_rook() & ~cp.white_queen());
    m_white_rook_attacks|=ats;
    m_white_attacks|=ats;
    ats&=~m_black_pawn_attacks;
    ats&=~m_black_knight_attacks;
    ats&=~m_black_bishop_attacks;
    ats&=~cp.white_pawn();
    m_safe_mobility+=rook_mobility[ats.popcount()];
  }
}

void Mobility::compute_black_rook_attacks()
{
  m_black_rook_attacks.clear();
  BitBoard bb{cp.black_rook()};
  while(bb){
    BitBoard s=bb.get_reset_square();
    BitBoard ats=m_atta.rook(s,cp.pieces() & ~cp.black_rook() & ~cp.black_queen());
    m_black_rook_attacks|=ats;
    m_black_attacks|=ats;
    ats&=~m_white_pawn_attacks;
    ats&=~m_white_knight_attacks;
    ats&=~m_white_bishop_attacks;
    ats&=~cp.black_pawn();
    m_safe_mobility-=rook_mobility[ats.popcount()];
  }
}

void Mobility::compute_white_queen_attacks()
{
  m_white_queen_attacks.clear();
  BitBoard bb{cp.white_queen()};
  while(bb){
    BitBoard s=bb.get_reset_square();
    BitBoard ats=m_atta.queen(s,cp.pieces());
    m_white_queen_attacks|=ats;
    m_white_attacks|=ats;
    ats&=~m_black_pawn_attacks;
    ats&=~m_black_knight_attacks;
    ats&=~m_black_bishop_attacks;
    ats&=~cp.white_pawn();
    m_safe_mobility+=queen_mobility[ats.popcount()];
  }
}

void Mobility::compute_black_queen_attacks()
{
  m_black_queen_attacks.clear();
  BitBoard bb{cp.black_queen()};
  while(bb){
    BitBoard s=bb.get_reset_square();
    BitBoard ats=m_atta.queen(s,cp.pieces());
    m_black_queen_attacks|=ats;
    m_black_attacks|=ats;
    ats&=~m_white_pawn_attacks;
    ats&=~m_white_knight_attacks;
    ats&=~m_white_bishop_attacks;
    ats&=~cp.black_pawn();
    m_safe_mobility-=queen_mobility[ats.popcount()];
  }
}

void Mobility::compute_white_king_attacks()
{
  m_white_king_attacks.clear();
  BitBoard m_white_king_attacks=m_atta.king(cp.white_king());
  m_white_attacks|=m_white_king_attacks;
}

void Mobility::compute_black_king_attacks()
{
  m_black_king_attacks.clear();
  BitBoard m_black_king_attacks=m_atta.king(cp.black_king());
  m_black_attacks|=m_black_king_attacks;
}

void Mobility::compute()
{
  m_safe_mobility=0;
  
  compute_white_king_neighbourhood();
  compute_black_king_neighbourhood();

  // pawns:
  compute_white_pawn_attacks();
  compute_black_pawn_attacks();

  // first minor pieces: 
  compute_white_knight_attacks();
  compute_black_knight_attacks();
  compute_white_bishop_attacks();
  compute_black_bishop_attacks();

  // rooks:
  compute_white_rook_attacks();
  compute_black_rook_attacks();

  // queen:
  compute_white_queen_attacks();
  compute_black_queen_attacks();

  // king:
  compute_white_king_attacks();
  compute_black_king_attacks();
 
}

int Mobility::safe_mobility()const
{
  return m_safe_mobility;
}

const BitBoard& Mobility::white_attacks()const
{
  return m_white_attacks;
}

const BitBoard& Mobility::black_attacks()const
{
  return m_black_attacks;
}

int Mobility::undefended_pieces()const
{
  return (cp.black_pieces() & ~m_black_attacks).popcount() - (cp.white_pieces() & ~m_white_attacks).popcount();
}

int Mobility::hanging_pieces()const
{
  BitBoard white_hanging=(cp.white_pieces() & ~m_white_attacks) & m_black_attacks;
  BitBoard black_hanging=(cp.black_pieces() & ~m_black_attacks) & m_white_attacks;
  white_hanging|=cp.white_knight() & m_black_pawn_attacks;
  white_hanging|=cp.white_bishop() & m_black_pawn_attacks;
  white_hanging|=cp.white_rook() & (m_black_pawn_attacks | m_black_knight_attacks | m_black_bishop_attacks);
  white_hanging|=cp.white_queen() & (m_black_pawn_attacks | m_black_knight_attacks | m_black_bishop_attacks | m_black_rook_attacks);
  black_hanging|=cp.black_knight() & m_white_pawn_attacks;
  black_hanging|=cp.black_bishop() & m_white_pawn_attacks;
  black_hanging|=cp.black_rook() & (m_white_pawn_attacks | m_white_knight_attacks | m_white_bishop_attacks);
  black_hanging|=cp.black_queen() & (m_white_pawn_attacks | m_white_knight_attacks | m_white_bishop_attacks | m_white_rook_attacks);
  return black_hanging.popcount() - white_hanging.popcount();
}



//------------------------------ king safety: -----------------------
//
//
//
void Mobility::compute_white_king_neighbourhood()
{
  BitBoard bb{cp.white_king()};
  bb|=bb<<8;
  bb|=bb>>8;
  bb|=bb<<1 & not_a_file;
  bb|=bb>>1 & not_h_file;  
  m_white_king_neighbourhood=bb;
  m_white_king_battlefield=bb | bb<<16 | bb<<24;
}

void Mobility::compute_black_king_neighbourhood()
{
  BitBoard bb{cp.black_king()};
  bb|=bb<<8;
  bb|=bb>>8;
  bb|=bb<<1 & not_a_file;  
  bb|=bb>>1 & not_h_file;  
  m_black_king_neighbourhood=bb;  
  m_black_king_battlefield=bb | bb>>16 | bb>>24;
}

const BitBoard& Mobility::white_king_neighbourhood()const
{
  return m_white_king_neighbourhood;
}

const BitBoard& Mobility::black_king_neighbourhood()const
{
  return m_black_king_neighbourhood;
}

const BitBoard& Mobility::white_king_battlefield()const
{
  return m_white_king_battlefield;
}

const BitBoard& Mobility::black_king_battlefield()const
{
  return m_black_king_battlefield;
}

int Mobility::attackweight_white_king()const
{
  int r{0};
  if(m_black_pawn_attacks & m_white_king_neighbourhood){r++;}
  if(m_black_knight_attacks & m_white_king_neighbourhood){r++;}
  if(m_black_bishop_attacks & m_white_king_neighbourhood){r++;}
  if(m_black_rook_attacks & m_white_king_neighbourhood){r++;}
  if(m_black_queen_attacks & m_white_king_neighbourhood){r++;}
  return m_attackweight[r];
}

int Mobility::attacks_on_white_king()const
{
  int r{0};
  r+=10*(m_black_pawn_attacks & m_white_king_neighbourhood).popcount();
  r+=20*(m_black_knight_attacks & m_white_king_neighbourhood).popcount();
  r+=20*(m_black_bishop_attacks & m_white_king_neighbourhood).popcount();
  r+=40*(m_black_rook_attacks & m_white_king_neighbourhood).popcount();
  r+=80*(m_black_queen_attacks & m_white_king_neighbourhood).popcount();
  return r;
}

int Mobility::attackweight_black_king()const
{
  int r{0};
  if(m_white_pawn_attacks & m_black_king_neighbourhood){r++;}
  if(m_white_knight_attacks & m_black_king_neighbourhood){r++;}
  if(m_white_bishop_attacks & m_black_king_neighbourhood){r++;}
  if(m_white_rook_attacks & m_black_king_neighbourhood){r++;}
  if(m_white_queen_attacks & m_black_king_neighbourhood){r++;}
  return m_attackweight[r];
}

int Mobility::attacks_on_black_king()const
{
  int r{0};
  r+=10*(m_white_pawn_attacks & m_black_king_neighbourhood).popcount();
  r+=20*(m_white_knight_attacks & m_black_king_neighbourhood).popcount();
  r+=20*(m_white_bishop_attacks & m_black_king_neighbourhood).popcount();
  r+=40*(m_white_rook_attacks & m_black_king_neighbourhood).popcount();
  r+=80*(m_white_queen_attacks & m_black_king_neighbourhood).popcount();
  return r;
}

int Mobility::king_safety()const
{
  return attackweight_black_king() * attacks_on_black_king() - attackweight_white_king() * attacks_on_white_king();
}
