/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"../board/cposition.h"
#include"../common/bitboard.h"
#include"move.h"
#include<iostream>
#include<sstream>
using std::cout;
using std::endl;
using std::ostringstream;


Move::Move()
{
  m_move=0;
}

Move::Move(Square a,Square b,uint32_t c,Piece p1,Piece p2)
{
  m_move=b.id()&0x3f;
  m_move|=(a.id()&0x3f)<<6;
  m_move|=c<<12;
  m_move|=p2.id()<<16;
  m_move|=p1.id()<<20;
  m_move&=0x00ffffff;
  m_move|=xnormal_move << 24;
}

Move::Move(Square a,Square b,uint32_t c,piece_t p1,piece_t p2)
{
  m_move=b.id()&0x3f;
  m_move|=(a.id()&0x3f)<<6;
  m_move|=c<<12;
  m_move|=p2<<16;
  m_move|=p1<<20;
  m_move&=0x00ffffff;
  m_move|=xnormal_move << 24;
}

bool Move::operator==(Move x)const
{
  return (m_move&0x00ffffff)==(x.m_move&0x00ffffff);
}

bool Move::operator!=(Move x)const
{
  return (m_move&0x00ffffff)!=(x.m_move&0x00ffffff);
}

string Move::algebraic(bool chess960)const
{
  string r;
  if(m_move==0){r="0000";return r;}
  r=from().file();
  r+=from().rank();
  r+=to().file();
  r+=to().rank();
  switch(code())
    {
    case normal_move:break;
    case promo_queen:r+='q';break;
    case promo_rook:r+='r';break;
    case promo_bishop:r+='b';break;
    case promo_knight:r+='n';break;
    case castling:
      if(!chess960){
	if(r=="e1h1"){r="e1g1";}
	else if(r=="e1a1"){r="e1c1";}
	else if(r=="e8h8"){r="e8g8";}
	else if(r=="e8a8"){r="e8c8";}
      }
      break;
    case en_passant:break;
    case nullmove:break;
    }
  return r;
}

string Move::dump()const
{
  ostringstream oss;
  oss << "from:" << from().name() << "  ";
  oss << "to:" << to().name() << "  ";
  oss << "code:";
  switch(code())
    {
    case nullmove:oss << "nullmove  ";break;
    case normal_move:oss << "normal_move  ";break;
    case promo_queen:oss << "promo_queen  ";break;
    case promo_rook:oss << "promo_rook  ";break;
    case promo_bishop:oss << "promo_bishop  ";break;
    case promo_knight:oss << "promo_knight  ";break;
    case castling:oss << "castling  ";break;
    case en_passant:oss << "en_passant  ";break;
    }
  oss << "piece:" << piece().name() << "  ";
  oss << "capture:" << capture().name() << "  ";
  oss << "check:" << (check()?"true":"false") << "   ";
  oss << "xcode:";
  switch(xcode())
    {
    case hashtable_move:oss << "hashtable_move" << endl;break;
    case good_capture:oss << "good_capture" << endl;break;
    case killer:oss << "killer" << endl;break;
    case killer_2:oss << "killer_2" << endl;break;
    case counter_move:oss << "counter_move" << endl;break;
    case losing_capture:oss << "losing_capture" << endl;break;
    case historical_move:oss << "historical_move" << endl;break;
    case fertility1:oss << "fertility1" << endl;break;
    case fertility2:oss << "fertility2" << endl;break;
    case fertility3:oss << "fertility3" << endl;break;
    case to_centre:oss << "to_centre" << endl;break;
    case to_bigcentre:oss << "to_bigcentre" << endl;break;
    case xnormal_move:oss << "xnormal_move" << endl;break;
    }
  return oss.str();
}

Square Move::from()const
{
  if(m_move==0){return no_square;}
  Square r=m_move>>6 & 0x3f;
  return r;
}

void Move::from(const Square& x)
{
  constexpr uint32_t a=0x3f;
  constexpr uint32_t b=~(a<<6);
  m_move&=b;
  m_move|=x.id()<<6;
  return;
}

Square Move::to()const
{
  if(m_move==0){return no_square;}
  Square r=m_move & 0x3f;
  return r;
}

void Move::to(const Square& x)
{
  constexpr uint32_t a=0x3f;
  constexpr uint32_t b=~a;
  m_move&=b;
  m_move|=x.id()&0x3f;
  return;
}

move_t Move::code()const
{
  int r=m_move>>12 & 0x7;
  return static_cast<move_t>(r);
}

void Move::code(const move_t& x)
{
  m_move&=0xffff8fff;
  m_move|=static_cast<int>(x)<<12;
  return;
}

xmove_t Move::xcode()const
{
  unsigned int r=m_move>>24 & 0xff;
  return static_cast<xmove_t>(r);
}

void Move::xcode(const xmove_t& x)
{
  m_move&=0x00ffffff;
  m_move|=static_cast<uint32_t>(x)<<24;
  return;
}

void Move::piece_id(int x)
{
  constexpr uint32_t a=0xf;
  constexpr uint32_t b=~(a<<20);
  m_move&=b;
  m_move|=static_cast<uint32_t>(x)<<20;
  return;
}

int Move::piece_id()const
{
  return (m_move>>20 & 0xf);
}

Piece Move::piece()const
{
  Piece r;
  r.set_id(m_move>>20 & 0xf);
  return r;
}

void Move::piece(const Piece& x)
{
  constexpr uint32_t a=0xf;
  constexpr uint32_t b=~(a<<20);
  m_move&=b;
  m_move|=static_cast<uint32_t>(x.id())<<20;
  return;
}

void Move::piece(uint32_t&& x)
{
  constexpr uint32_t a=0x0f;
  constexpr uint32_t b=~(a<<20);
  m_move&=b;
  m_move|=x<<20;
  return;
}

Piece Move::capture()const
{
  Piece r;
  r.set_id(m_move>>16 & 0xf);
  return r;
}

int Move::capture_id()const
{
  return (m_move>>16 & 0xf);
}

void Move::capture_id(int x)
{
  constexpr uint32_t a=0xf;
  constexpr uint32_t b=~(a<<16);
  m_move&=b;
  m_move|=static_cast<uint32_t>(x)<<16;
  return;  
}

void Move::capture(const Piece& x)
{
  constexpr uint32_t a=0xf;
  constexpr uint32_t b=~(a<<16);
  m_move&=b;
  m_move|=static_cast<uint32_t>(x.id())<<16;
  return;
}

void Move::capture(uint32_t&& x)
{
  constexpr uint32_t a=0xf;
  constexpr uint32_t b=~(a<<16);
  m_move&=b;
  m_move|=x<<16;
  return;
}

bool Move::check()const
{
  return m_move & 0x8000;
}

void Move::check(bool x)
{
  m_move&=0xffff7fff;
  m_move|=static_cast<int>(x)<<15;
}

void Move::clear()
{
  m_move=0;
}
