/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"../board/cposition.h"
#include"../eval/evaluation.h"
#include"killer.h"
#include"../hashtable/transposition.h"
#include"../movegen/movelist.h"
#include"analyze.h"
#include"../common/numerals.h"


#include<algorithm>
#include<thread>
#include<ratio>
#include<chrono>
#include<stdexcept>
#include<iostream>
#include<iomanip>
#include<utility>
#include<sstream>

using std::to_string;
using std::back_inserter;
using std::hex;
using std::dec;
using std::cout;
using std::flush;
using std::endl;
using std::pair;
using std::boolalpha;
using std::logic_error;
using std::istringstream;
using std::ostringstream;
using std::setprecision;
using std::setw;
using namespace std::chrono;



Analysis::Analysis(ChessPosition& x,TranspositionTable& y,ThreadPool& tp): cp(x), m_tt(y), m_tp(tp), m_eva(x,m_egtkbnk,m_egtkqkr)
{
  m_xml.set_analysis(this);
  m_ply=0;
  m_max_depth=max_search_depth-1;
  m_sd_depth=m_max_depth;
  m_node_count=0;
  m_qnodes=0;
  m_max_nodes=infinite_nodes;
  m_action=no_action;
  m_status=idle;
  m_date=0;
  m_quit=false;
  init(cp);
  m_tt.clear();

}

Analysis::~Analysis()
{
}

Analysis& Analysis::operator=(const Analysis& x)
{
  cp=x.cp;
  m_xml.set_analysis(this);
  m_eva.init();
  m_action=no_action;
  m_status=idle;
  m_quit=false;
  m_tman.set_infinite(true);
  
  m_iteration=x.m_iteration;
  m_ply=x.m_ply;
  m_seldepth=x.m_seldepth;
  m_nullmove_level=x.m_nullmove_level;
  m_date=x.m_date;
  m_max_depth=x.m_max_depth;
  m_sd_depth=x.m_sd_depth;
  m_max_nodes=x.m_max_nodes;
  
  m_node_count=x.m_node_count;
  m_qnodes=x.m_qnodes;
  m_total_node_count=0;
  m_total=0;
  m_leave_nodes=0;
  m_iid_sort=0;
  m_ebf=0;
  
  return *this;
}

void Analysis::init(const ChessPosition& x)
{
  if(m_status==thinking){
    std::cerr << "init() while thinking not allowed" << endl;
    throw logic_error("Analysis::init  init() call while thinking not allowed");
  }
  cp=x;
  m_action=no_action;
  m_status=idle;
  m_eva.init();
  m_tman.reset();
}

void Analysis::reset()
{
  m_sd_depth=m_max_depth;
  m_searchmoves.clear();
  m_tman.reset();
}

void Analysis::set_wtime(int x)
{
  m_tman.set_wtime(x);
}

void Analysis::set_btime(int x)
{
  m_tman.set_btime(x);
}

void Analysis::set_winc(int x)
{
  m_tman.set_winc(x);
}

void Analysis::set_binc(int x)
{
  m_tman.set_binc(x);
}

void Analysis::set_movestogo(int x)
{
  m_tman.set_white_to_move(cp.white_to_move());
  m_tman.set_movestogo(x);
}

void Analysis::set_depth(int x)
{
  m_sd_depth=x;
}

void Analysis::set_nodes(long long x)
{
  m_max_nodes=x;
  m_tman.set_nodes(x);
}

void Analysis::set_mate(int x)
{
  m_sd_depth=x;
}

void Analysis::set_movetime(int x)
{
  m_tman.set_movetime(x);
}

void Analysis::set_infinite(bool x)
{
  m_tman.set_infinite(x);
}

void Analysis::set_pondering(bool x)
{
  if(x){m_tman.set_pondering(true);}
  else{m_tman.set_pondering(false);}
}

void Analysis::set_searchmoves(vector<Move> x)
{
  m_tt.clear();
  m_searchmoves=x;
}

void Analysis::set_uci_limitstrength(bool x)
{
  m_uci_limitstrength=x;
  if(!x){
    m_elo_maxdepth=max_search_depth;
    m_handicap=0.0;
  }
}

void Analysis::set_uci_elo(int x)
{
  if(x<1500){x=1500;}
  if(x>2200){x=2200;}
  m_uci_elo=x;
  if(x<1600){m_elo_maxdepth=2;}
  else if(x<1700){m_elo_maxdepth=4;}
  else if(x<1800){m_elo_maxdepth=6;}
  else if(x<1900){m_elo_maxdepth=8;}
  else if(x<2000){m_elo_maxdepth=10;}
  else if(x<2100){m_elo_maxdepth=12;}
  else if(x<2200){m_elo_maxdepth=14;}
  if(x>=2000){m_handicap=0;}
  if(x<2000){m_handicap=200*(2000-x)/500;}
}

void Analysis::clear_eval()
{
  m_eva.clear();
}

void Analysis::clear_hash()
{
  m_tt.clear();
}

void Analysis::hashtable_enabled(bool x)
{
  m_tt.set_enabled(x);
}

void Analysis::clear_killer()
{
  m_killer.clear();
}

void Analysis::clear_history()
{
  m_ht.clear();
}

void Analysis::clear_countermove()
{
  m_cm.clear();
}

void Analysis::set_log(bool x)
{
  m_log=x;
  ostringstream oss;
  oss << "analysis_" << cp.id() << ".log";
  if(x){
    ofs_log.open(oss.str());
  }else{
    ofs_log.close();
  }
  m_eva.set_log(x);
}

void Analysis::quit()
{
  m_quit=true;
}

void Analysis::set_action(action_t x)
{
  m_action=x;
}

int Analysis::get_status()const
{
  return m_status;
}

void Analysis::set_status(int x)
{
  m_status=x;
}

void Analysis::set_thread_depth(float x)
{
  m_thread_depth=x;
}

void Analysis::set_thread_alpha(int x)
{
  m_thread_alpha=x;
}

void Analysis::set_thread_beta(int x)
{
  m_thread_beta=x;
}

void Analysis::set_thread_lmr(int x)
{
  m_thread_lmr=x;
}

void Analysis::set_thread_reductions(int x)
{
  m_thread_reductions=x;
}

void Analysis::set_thread_extensions(int x)
{
  m_thread_extensions=x;
}

const int& Analysis::get_thread_score()const
{
  return m_thread_score;
}

const long long& Analysis::get_thread_nodes()const
{
  return m_thread_nodes;
}

Duration Analysis::get_thread_time()
{
  return m_thread_timer.total_time();
}

const int& Analysis::get_thread_id()const
{
  return m_thread_id;
}

void Analysis::set_thread_id(int x)
{
  m_thread_id=x;
}

void Analysis::ponderhit()
{
  m_tman.set_pondering(false);
}

void Analysis::new_node()
{
  m_node_count++;
  m_thread_nodes++;
  if(m_max_nodes!=infinite_nodes){
    if(m_node_count>m_max_nodes){
      m_action=stop_thinking;
    }
  }else if((m_node_count&0x3FFF)==0){
    if(m_ply>2 && m_tman.time_to_move(m_node_count)){
      m_action=stop_thinking;
    }
  }
}

int Analysis::root_search(float depth,int alpha,int beta)
{
  //------------------------- EGT ? ----------------------------------------------------------------------
  {
    bool egt{false};
    if(!cp.white_pawn() && !cp.black_pawn()){
      if((cp.white_queen() | cp.black_queen()
	  | cp.white_rook() | cp.black_rook()
	  | cp.white_bishop() | cp.black_bishop()
	  | cp.white_knight() | cp.black_knight()).popcount()<=2){      
	if(cp.materialindex().str()=="KBNk" || cp.materialindex().str()=="Kkbn"){
	  //
	  egt=true;
	  if(!m_egtkbnk.ready()){
	    m_tt.set_threshold(6000000);
	    if(!m_egtkbnk.busy()){m_egtkbnk.set_memory(m_tt.get_uchar_adress());}
	  }
	  while(m_tman.remaining_time()>2000){
	    m_egtkbnk.create();
	    if(m_egtkbnk.ready()){
	      break;
	    }
	  }
	  //
	}else if(cp.materialindex().str()=="KQkr" || cp.materialindex().str()=="KRkq" || cp.materialindex().str()=="KQk"
		 || cp.materialindex().str()=="Kkq" || cp.materialindex().str()=="KRk" || cp.materialindex().str()=="Kkr"){
	  //
	  egt=true;
	  if(!m_egtkqkr.ready()){
	    m_tt.set_threshold(6000000);
	    if(!m_egtkqkr.busy()){
	      m_egtkqkr.set_memory(m_tt.get_uchar_adress());
	    }
	  }
	  while(m_tman.remaining_time()>4000){
	    m_egtkqkr.create();
	    if(m_egtkqkr.ready()){
	      break;
	    }
	  }
	  //
	}
      }
    }
    if(!egt){
      m_egtkqkr.clear();
      m_egtkbnk.clear();
      m_tt.set_threshold(0);
    }
  }
  //------------------------------------------------------------------------------------------------------

  PrincipalVariation pv;
  int eval=pv_search(depth,alpha,beta,pv);
  m_mutex_mpv.lock();
  m_mpv.clear();
  m_mpv.add(m_pv);
  m_mutex_mpv.unlock();

  if(m_mpv.number()>1){
    auto old_value=m_value;
    auto old_best_move=m_best_move;

    PrincipalVariation best_pv;
    Move best_move;
    MList ml;
    cp.generate_moves(ml);
    ml.remove(m_best_move);
    for(int i=1;i<m_mpv.number();i++){  
      alpha=-infinite_value;
      beta=infinite_value;
      for(size_t j=0;j<ml.size();j++){
	make_move(ml[j]);
	int v=-pv_search(depth-1,alpha,beta,pv);
	unmake_move(ml[j]);
	if(v>alpha){
	  alpha=v;
	  best_move=ml[j];
	  best_pv=ml[j]+pv;
	  best_pv.set_score(alpha);
	  best_pv.set_depth(depth);
	}
      }
      m_mutex_mpv.lock();
      m_mpv.add(best_pv);
      m_mutex_mpv.unlock();
      ml.remove(best_move);      
    }
    
    m_value=old_value;
    m_best_move=old_best_move;
  }
  
  return eval;
}

int Analysis::pv_search(const float& depth,int alpha,const int& beta,PrincipalVariation& pv)
{
  if(cp.draw_by_threefold_repetition() && m_ply>0){return 0;}
  if(cp.get_dfz()>100 && m_ply>0){return 0;}
  //---------------------- egt pruning ----------------------------
  //
  if(m_ply>0){
    int static_value=static_eval();
    if(static_value>mate_value-max_search_depth || static_value<-mate_value+max_search_depth){
      return static_value;
    }
  }
  //
  //---------------------------------------------------------------
  pv.clear();
  if(m_ply>m_seldepth){m_seldepth=m_ply;}
  if(m_action==stop_thinking){return alpha;}
  bound_t cutoff=upperbound;
  PrincipalVariation child_pv;
  Move best_move;
  int value;
  if(depth<=0 || m_ply>=max_search_depth){
    value=quiescence(alpha,beta);
    if(cp.get_dfz()>50){
      float f=1.0-0.02*(cp.get_dfz()-50.0);
      value*=f;
    }
    m_leave_nodes++;
    return value; 
  }else{
    //----------------------- hashtable move ------------------------
    //
    TTrecord tt;
    tt.set_key(cp.get_hash());
    if(m_tt.read(tt)){
      TTmove ttm=tt.get_move();
      if(is_legal(ttm)){
	best_move=ttm(cp);
      }
    }
    //
    //----------------------- hashtable move ------------------------
    
    BitBoard passed_pawns=m_eva.passed_pawns();
    BitBoard white_king_battlefield=m_eva.white_king_battlefield();
    BitBoard black_king_battlefield=m_eva.black_king_battlefield();

    MoveList ml(m_xml);
    cp.generate_moves(ml);
    select_searchmoves(ml);
    if(ml.size()==0){
      if(!cp.check()){return 0;}else{return -mate_value+m_ply;}
    }else{
      if(m_iteration==1 && m_ply==0){m_best_move=ml[0];}
    }
    if(best_move.code()!=nullmove){ml.set_hashtable_move(best_move);}
    ml.set_killers(m_killer.get_killerA(m_ply),m_killer.get_killerB(m_ply));
    ml.set_killers_2(m_killer.get_killerA(m_ply-2),m_killer.get_killerB(m_ply-2));    
    if(depth>4 && best_move.code()==nullmove){iidsort(ml,depth-3,alpha,beta);}
    ml.sort();
    if(m_ply==0 && alpha<pawn_value){
      sort_fertility(ml);
    }
    
    Move m;
    for(size_t mi=0;mi<ml.size();mi++){
      m=ml[mi];

      //------------------- LMR: --------------------------------------------
      //
      int lmr{0};	
      if(
	 !cp.check() && !m.capture() && !m.check() &&
	 m.xcode()!=killer && m.xcode()!=killer_2 &&
	 !m.piece().king()
	 )
	{
	  lmr=m_lmr.value(depth,mi);
	}
      //
      //---------------------------------------------------------------------
      
      //----------------- extensions: ---------------------------------------
      //
      int extension=0;
      if(m.check() && see(m)>=0){extension++;lmr=0;}  // check extension
      if(m_ply < 2*m_iteration){
	if(ml.size()==1){extension++;lmr=0;}                    // single move extension
	if(m.capture() && cp.history().size()>0 && cp.history().back().last_move.capture() && m.to()==cp.history().back().last_move.to()){
	  if((cp.white_to_move() && (m.to().bitboard() & black_king_battlefield)) ||
	     (!cp.white_to_move() && (m.to().bitboard() & white_king_battlefield))){
	    extension++;lmr=0;                          // recapture extension in king battlefield 
	  }
	}
	if(cp.nobelmen()<=2*rook_value){
	  if(m.piece().pawn() && (m.from().bitboard() & passed_pawns)){extension++;lmr=0;}  // passed pawn race extension
	  else if(m.piece().white_pawn() && m.to().rank_number()==7){extension++;lmr=0;}  // passed pawn race extension
	  else if(m.piece().black_pawn() && m.to().rank_number()==2){extension++;lmr=0;}  // passed pawn race extension
	}
      }
      //
      //---------------------------------------------------------------------
      
      uint64_t local_nodes=m_node_count;
      make_move(m);
      if(mi==0){               // principal variation search
	value=-pv_search(depth-1+extension,-beta,-alpha,child_pv);
      }else{
	value=-nullwindow_search(depth-1-lmr+extension,-alpha-1,-alpha,0,0);
	if(value>alpha){
	  if(lmr>0){           // repeat null window search without lmr
	    value=-nullwindow_search(depth-1+extension,-alpha-1,-alpha,0,0);
	    if(value>alpha){   // repeat search with open window
	      value=-pv_search(depth-1+extension,-beta,-alpha,child_pv);
	    }
	  }else{
	    value=-pv_search(depth-1+extension,-beta,-alpha,child_pv);
	  }
	}
      }
      unmake_move(m);
      if(m_action==stop_thinking){return alpha;}
      set_fertility(m,m_node_count-local_nodes);

      if(value>=beta){
	put_in_tt(lowerbound,depth,alpha,m,m_date);
	return beta;
      }
      if(value>alpha){
	best_move=m;
	alpha=value;
	cutoff=exact_value;
	pv.clear();
	pv=best_move+child_pv;
	if(m_ply==0){
	  pv.set_depth(depth);
	  pv.set_score(alpha);
	  m_pv=pv;
	  m_value=alpha;
	  m_best_move=best_move;
  	}
      }
    }
    
    put_in_tt(cutoff,depth,alpha,best_move,m_date);
  }

  if(m_ply==0 && alpha==-infinite_value){return 0;}
  return alpha;
}

int Analysis::nullwindow_search(const float& depth,int alpha,const int& beta,const int& reductions,const int& extensions)
{
  if(m_ply>m_seldepth){m_seldepth=m_ply;}
  if(m_action==stop_thinking){return alpha;}
  if(cp.draw_by_threefold_repetition()){return 0;}
  bound_t cutoff=upperbound;
  Move best_move,hashtable_move;
  int value;

  
  if(depth<=0 || m_ply>=max_search_depth)
    {
      m_leave_nodes++;
      return quiescence(alpha,beta); 
    }
  else
    {        
      int static_value=cp.check()?alpha:static_eval();
      
      //---------------------- egt pruning ----------------------------
      //
      if(static_value>mate_value-max_search_depth || static_value<-mate_value+max_search_depth){
	return static_value;
      }
      //
      //---------------------------------------------------------------

      
      BitBoard passed_pawns=m_eva.passed_pawns();
      BitBoard white_king_battlefield=m_eva.white_king_battlefield();
      BitBoard black_king_battlefield=m_eva.black_king_battlefield();
      
      //---------------------- nullmove pruning -----------------------
      //
      if(static_value>=beta && nullmove_allowed()){
	value=nullmove_search(depth,alpha,beta);
	if(value>=beta){
	  put_in_tt(lowerbound,depth,beta,best_move,m_date);
	  return beta;
	}
      }
      //
      //---------------------- nullmove pruning -----------------------
      


      //----------------------- hashtable move ------------------------
      //
      TTrecord tt;
      tt.set_key(cp.get_hash());
      if(m_tt.read(tt)){
	TTmove ttm=tt.get_move();
	auto score=tt.get_score(m_ply);
	if(tt.get_draft()>=depth){
	  if(tt.is_exact_value()){return score;}
	  else if(tt.is_upperbound() && score<=alpha){
	    m_cut_off_xmove_t[depth][0]++;
	    return alpha;
	  }
	  else if(tt.is_lowerbound() && score>=beta){
	    m_cut_off_xmove_t[depth][0]++;
	    return beta;
	  }
	}
	if(is_legal(ttm)){
	  best_move=ttm(cp);
	  
	  int extension{0};
	  if(best_move.check() && see(best_move)>=0){extension++;}  // check extension
	  
	  make_move(best_move);
	  value=-nullwindow_search(depth-1+extension,-beta,-alpha,reductions,extensions+extension);
	  unmake_move(best_move);
	  if(value>=beta){
	    m_cut_off_xmove_t[depth][0]++;
	    return beta;
	  }
	}
      }
      //
      //----------------------- hashtable move ------------------------

      MoveList ml(m_xml);
      cp.generate_moves(ml);
      if(ml.size()==0){if(!cp.check()){return 0;}else{return -mate_value+m_ply;}}
      if(best_move.code()!=nullmove){ml.set_hashtable_move(best_move);}
      ml.set_killers(m_killer.get_killerA(m_ply),m_killer.get_killerB(m_ply));
      ml.set_killers_2(m_killer.get_killerA(m_ply-2),m_killer.get_killerB(m_ply-2));
      ml.set_xcodes();
      
      Move m=ml.first();
      for(size_t mi=0;mi<ml.size();mi++){
	if(m==best_move){m=ml.next();continue;}
	
	//---------------------- LMP (late move pruning) ------------------
	//
	if(m_ply>4){
	  if((depth==1 && mi>6) || (depth==2 && mi>6) || (depth==3 && mi>10) || (depth==4 && mi>10)){
	    if(!m.capture() && m.code()==normal_move && !m.check()){
	      m=ml.next();continue;
	    }
	  }
	}
	//
	//---------------------------------------------------------------
	
	//------------------ futility pruning: --------------------------------
	//
	if(!m.capture() && m.code()==normal_move && !m.check()){
	  if((depth==1 && static_value <= alpha - pawn_value) ||
	     (depth==2 && static_value <= alpha - rook_value) ||
	     (depth==3 && static_value <= alpha - queen_value)){
	    m=ml.next();continue;
	  }
	}
	//
	//---------------------------------------------------------------------
	
	//------------------- LMR: --------------------------------------------
	//
	int lmr{0};
	if(
	   !cp.check() && !m.capture() && !m.check() &&
	   m.xcode()!=killer && m.xcode()!=killer_2 &&
	   !m.piece().king()
	   )
	  {
	    lmr=m_lmr.value(depth,mi);
	  }
	//
	//---------------------------------------------------------------------
	
	
	//----------------- extensions: ---------------------------------------
	//
	int extension=0;
	if(extensions-reductions<5){
	  if(m.check() && see(m)>=0){extension++;lmr=0;}  // check extension
	  if(m_ply < 2*m_iteration){
	    if(ml.size()==1){extension++;lmr=0;}                    // single move extension
	    if(m.capture() && cp.history().size()>0 && cp.history().back().last_move.capture() && m.to()==cp.history().back().last_move.to()){
	      if((cp.white_to_move() && (m.to().bitboard() & black_king_battlefield)) ||
		 (!cp.white_to_move() && (m.to().bitboard() & white_king_battlefield))){
		extension++;lmr=0;                          // recapture extension in king battlefield 
	      }
	    }
	    if(cp.nobelmen()<=2*rook_value){
	      if(m.piece().pawn() && (m.from().bitboard() & passed_pawns)){extension++;lmr=0;}  // passed pawn race extension
	      else if(m.piece().white_pawn() && m.to().rank_number()==7){extension++;lmr=0;}  // passed pawn race extension
	      else if(m.piece().black_pawn() && m.to().rank_number()==2){extension++;lmr=0;}  // passed pawn race extension
	    }
	  } 
	}
	//
	//---------------------------------------------------------------------
	
	make_move(m);
	
	value=-nullwindow_search(depth-1-lmr+extension,-beta,-alpha,lmr+reductions,extensions+extension);  // null window search with LMR
	if(value>=beta && lmr>0){
	  value=-nullwindow_search(depth-1,-beta,-alpha,lmr+reductions,extensions+extension);              // repeat search without LMR
	}
	
	unmake_move(m);
	if(m_action==stop_thinking){return alpha;}
	if(value>alpha){
	  best_move=m;
	  alpha=beta;
	  cutoff=lowerbound;
	  if(!best_move.capture()){
	    // countermove:
	    size_t hs{cp.history().size()};
	    if(hs>0 && cp.history().back().last_move!=NullMove()){
	      m_cm.increment(cp.history().back().last_move,best_move,depth);
	      for(unsigned i=0;ml[i]!=m;i++){
		m_cm.decrement(cp.history().back().last_move,ml[i],depth);
	      }
	      if(hs>1 && cp.history()[hs-2].last_move!=NullMove()){
		m_cm.increment(cp.history()[hs-2].last_move,best_move,depth);
		for(unsigned i=0;ml[i]!=m;i++){
		  m_cm.decrement(cp.history()[hs-2].last_move,ml[i],depth);
		}
	      }
	    }
	    // history table:
	    m_ht.increment(best_move,depth);
	    for(unsigned i=0;ml[i]!=m;i++){
	      m_ht.decrement(ml[i],depth);
	    }
	    // killers:
	    m_killer.set_killer(m_ply,best_move);
	  }
	  
	  break;
	}

	m=ml.next();
      }

      put_in_tt(cutoff,depth,alpha,best_move,m_date);
    }

  return alpha;
}


int Analysis::nullmove_search(float depth,int alpha,int beta)
{
  int R=depth/5+3;
  ++m_nullmove_level;
  make_nullmove();
  int value=-nullwindow_search(depth-R-1,-beta,-alpha,0,0);
  unmake_nullmove();
  --m_nullmove_level;
  return value;
}

void Analysis::iterative_search()
{
  m_eva.init();
  m_tman.start();
  m_tm.start();
  m_seldepth=0;
  m_nullmove_level=0;
  m_node_count=0;
  m_tp.clear_nodes();
  m_qnodes=0;
  m_leave_nodes=0;
  m_best_move.clear();
  m_pondermove.clear();
  m_value=-infinite_value;
  m_ht.divide_by_8();
  m_killer.clear();
  m_pv.clear();
  m_date++;
  int aspiration_alpha=-infinite_value;
  int aspiration_beta=infinite_value;
  m_iteration=1;
  while(m_iteration<m_max_depth && m_iteration<=m_sd_depth && m_value<mate_value && (!m_uci_limitstrength || m_iteration<=m_elo_maxdepth)){
    if(m_uci_limitstrength){m_tt.clear();}
    m_ply=0;
    m_nullmove_level=0;
    int alpha=aspiration_alpha,beta=aspiration_beta;
    int v=root_search(m_iteration,alpha,beta);
    if(m_mpv.size()>0){
      if(m_mpv[0].size()>0){m_best_move=m_mpv[0][0];}
      if(m_mpv[0].size()>1){m_pondermove=m_mpv[0][1];}else{m_pondermove.clear();}
    }
    
    //------------------------ aspiration window handling: --------------------------------------
    //
    if(m_baspiration && m_action!=stop_thinking && m_mpv.number()==1){
      if(v<=aspiration_alpha && aspiration_alpha!=-infinite_value){
	aspiration_alpha-=aspiration_beta-aspiration_alpha;
	if(aspiration_alpha<-rook_value){aspiration_alpha=-infinite_value;}
	continue;
      }else if(v>=aspiration_beta && aspiration_beta!=infinite_value){
	aspiration_beta+=aspiration_beta-aspiration_alpha;
	if(aspiration_beta>rook_value){aspiration_beta=infinite_value;}
	continue;
      }else{
	aspiration_alpha=v-pawn_value/4;
	aspiration_beta=v+pawn_value/4;
      }
    }
    //
    //------------------------ aspiration window handling  --------------------------------------
    
    
    if(m_action!=stop_thinking){
      float exponent=1.0/m_iteration;
      m_ebf=pow(m_leave_nodes,exponent);
    }
    m_leave_nodes=0;

    m_iteration++;
    if(m_action==stop_thinking){break;}
    //-------------------- end iteration ---------------------------------
  }

  if(m_uci_limitstrength){
    m_mpv.randomize(m_handicap);
    m_best_move=m_mpv[0][0];
    m_pondermove=NullMove();
    while(m_action!=stop_thinking && !m_tman.time_to_move(0)){
      std::this_thread::sleep_for(milliseconds(50));      
    }
  }
  
  m_time=m_tm.running_time().centiseconds();
  m_total_node_count+=m_tp.get_nodes();
  m_tm.stop();
  m_tp.stop_spinning();
  m_total++;
}

bool Analysis::nullmove_allowed()const
{
  if(cp.last_move().code()==nullmove){return false;}
  if(cp.check()){return false;}
  if(m_ply<1){return false;}
  if(cp.white_to_move()){
    if(cp.white_pieces() & ~cp.white_king() & ~cp.white_knight() & ~cp.white_pawn()){return true;}
    else{return false;}
  }else{
   if(cp.black_pieces() & ~cp.black_king() & ~cp.black_knight() & ~cp.black_pawn()){return true;}
    else{return false;}
  }
}

bool Analysis::is_legal(TTmove& m)const
{
  if(m.from()==m.to())return false;
  if(cp.white_to_move()){
    if(!(m.from().bitboard()&cp.white_pieces())){return false;}
    if(m.to().bitboard()&cp.black_king()){return false;}
  }else{
    if(!(m.from().bitboard()&cp.black_pieces())){return false;}
    if(m.to().bitboard()&cp.white_king()){return false;}
  }
  return true;
}

void Analysis::set_multipv(unsigned int n)
{
  if(n<1){n=1;}
  m_mpv.set_number(n);
}

void Analysis::set_max_nodes(long long n)
{
  m_max_nodes=n;
}

void Analysis::reset_max_nodes()
{
  m_max_nodes=infinite_nodes;
}

int Analysis::eval()const
{
  return m_value;
}

long long Analysis::get_node_count()const
{
  return m_node_count;
}

void Analysis::clear_node_count()
{
  m_node_count=0;
}

int Analysis::get_time()const
{
  return m_time;
}

int Analysis::get_running_time()const
{
  return m_tm.running_time().milliseconds();
}

float Analysis::get_ebf()const
{
  return m_ebf;
}

int Analysis::get_hashfull()const
{
  return m_tt.hashfull(m_date);
}

int Analysis::get_iteration()const
{
  return m_iteration;
}

int Analysis::get_seldepth()const
{
  return m_seldepth;
}

void Analysis::set_hash_size(int i)
{
  m_tt.set_size(i);
}

void Analysis::set_aspiration(bool x)
{
  m_baspiration=x;
}

void Analysis::select_searchmoves(MoveList& x)
{
  if(m_ply==0 && m_searchmoves.size()>0){
    x.set_searchmoves(m_searchmoves);
  }
}

const ChessPosition& Analysis::chessposition()const
{
  return cp;
}
PrincipalVariation Analysis::pv()const
{
  m_mutex_mpv.lock();
  PrincipalVariation r{m_mpv[0]};
  m_mutex_mpv.unlock();
  return r;
}

MultiPV Analysis::mpv()const
{
  m_mutex_mpv.lock();
  MultiPV r{m_mpv};
  m_mutex_mpv.unlock();
  return r;
}

Move Analysis::best_move()const
{
  return m_best_move;
}

Move Analysis::pondermove()const
{
  return m_pondermove;
}

void Analysis::make_move(const Move& m)
{
  cp.make_move(m);
  m_eva.make_move(m);
  new_node();
  ++m_ply;
}

void Analysis::unmake_move(const Move& m)
{
  m_eva.unmake_move(m);
  cp.unmake_move();
  --m_ply;
}

void Analysis::make_quies_move(const Move& m)
{
  cp.make_quies_move(m);
  m_eva.make_move(m);
  new_node();
  ++m_qnodes;
  ++m_ply;
}

void Analysis::unmake_quies_move(const Move& m)
{
  m_eva.unmake_move(m);
  cp.unmake_quies_move(m);
  --m_ply;
}

void Analysis::make_nullmove()
{
  NullMove m;
  cp.make_move(m);
  new_node();
  ++m_ply;
}

void Analysis::unmake_nullmove()
{
  cp.unmake_move();--m_ply;
}

void Analysis::operator()()
{
  while(true)
    {
      if(m_action==start_iterations){
	m_action=no_action;
	m_status=thinking;
	m_thread_timer.start();
	iterative_search();
	m_thread_timer.stop();
	m_status=finished;
      }else if(m_action==start_nws){
	m_action=no_action;
	m_status=thinking;
	m_node_count=0;
	
	m_thread_timer.start();
	m_thread_score=nullwindow_search(m_thread_depth,m_thread_alpha,m_thread_beta,m_thread_reductions,m_thread_extensions);
	m_thread_timer.stop();
	m_status=finished;
      }else if(m_action==stop_thinking){
	m_action=no_action;
	m_status=finished;
      }
      if(m_quit){break;}
      if(m_status!=spinning){
	std::this_thread::sleep_for(milliseconds(5));
      }
    }
}

const CounterMove& Analysis::countermove()const
{
  return m_cm;
}

const HistoryTable& Analysis::historytable()const
{
  return m_ht;
}

Evaluation& Analysis::evaluation()
{
  return m_eva;
}

void Analysis::iidsort(MoveList& ml,int d,int alpha,int beta)
{
  m_iid_sort++;
  PrincipalVariation null_pv;
  Move best_move;
  int score,best_score{-infinite_value};
  
  for(unsigned i=0;i<ml.size();i++){
    make_move(ml[i]);
    score=-pv_search(d,-beta,-alpha,null_pv);
    if(score>best_score){best_score=score;best_move=ml[i];}
    unmake_move(ml[i]);
  }
  ml.set_hashtable_move(best_move);  
}

int Analysis::see(Square s2)
{
  Square s1;
  switch(cp.stm()){
  case 0:s1=cp.black_weakest_attacker(s2);break;
  case 1:s1=cp.white_weakest_attacker(s2);break;
  default:throw logic_error("Analysis::see   side to move is corrupt");
  }

  if(s1==no_square){return 0;}
  Move m=cp.convert_to_move(s1,s2);
  if(m.capture().king()){return infinite_value;}
  int capval=m.capture().abs_value();
  switch(m.code())
    {
    case normal_move:break;
    case promo_queen:capval+=queen_value-pawn_value;break;
    case promo_rook:capval+=rook_value-pawn_value;break;
    case promo_bishop:capval+=bishop_value-pawn_value;break;
    case promo_knight:capval+=knight_value-pawn_value;break;
    case castling:break;
    case en_passant:break;
    case nullmove: break;
    }
  cp.force_move(m);
  int value=capval-see(s2);
  cp.undo_forced_move(m);

  if(value>0){return value;}else{return 0;}
}

int Analysis::see(const Move& m)
{
  int value;
  int capval=m.capture().abs_value();
  switch(m.code())
    {
    case normal_move:break;
    case promo_queen:capval+=queen_value-pawn_value;break;
    case promo_rook:capval+=rook_value-pawn_value;break;
    case promo_bishop:capval+=bishop_value-pawn_value;break;
    case promo_knight:capval+=knight_value-pawn_value;break;
    case castling:break;
    case en_passant:break;
    case nullmove: break;
    }
  Square tof=m.to();
  cp.force_move(m);
  value=capval-see(tof);
  cp.undo_forced_move(m);
  return value;
}

int Analysis::quiescence(int alpha,int beta,int depth)
{
  int value;

  int standpat=m_eva.static_eval();
  if(standpat>=mate_value-max_search_depth){standpat-=m_ply;}
  else if(standpat<=-mate_value+max_search_depth){standpat+=m_ply;}

  if(standpat>=beta){alpha=beta;return alpha;}
  if(alpha<standpat){alpha=standpat;}
  if(m_action==stop_thinking){return alpha;}

  MoveList ml(m_xml);
  if(cp.check()){
    cp.generate_moves(ml);
    if(ml.size()==0){alpha=-mate_value+m_ply;return alpha;}
  }else{
    cp.generate_quiescence_moves(ml);
  }
  for(unsigned i=0;i<ml.size();i++){
    auto& m=ml[i];
    if(depth<-8){
      if(m.piece().abs_value()<=m.capture().abs_value() || see(m)>0){
	make_quies_move(m);
	value=-quiescence(-beta,-alpha,depth-1);
	if(value>=mate_value-max_search_depth){value-=m_ply;}
	else if(value<=-mate_value+max_search_depth){value+=m_ply;}
	unmake_quies_move(m);
       	if(value>=beta){alpha=beta;return alpha;}
	if(value>alpha){alpha=value;}
      }
    }else if(m.piece().abs_value()<=m.capture().abs_value() || see(m)>=0){
      make_quies_move(m);
      value=-quiescence(-beta,-alpha,depth-1);
      if(value>=mate_value-max_search_depth){value-=m_ply;}
      else if(value<=-mate_value+max_search_depth){value+=m_ply;}
      unmake_quies_move(m);
      if(value>=beta){alpha=beta;return alpha;}
      if(value>alpha){alpha=value;}
    }
  }
  return alpha; 
}

int Analysis::static_eval()
{
  return m_eva.static_eval();
}

string Analysis::static_eval_report()
{
  return m_eva.report();
}

string Analysis::dump_line()
{
  ostringstream oss;
  oss << "(";
  oss << cp.get_make_move_counter();
  oss << ") ";
  oss << cp.history_str();
  oss << endl;
  return oss.str();
}

void Analysis::put_in_tt(bound_t co,int depth,int score,Move best_move,unsigned char date)
{
  TTrecord ttr;
  ttr.set_bound(co);
  ttr.set_draft(depth);
  ttr.set_score(score,m_ply);
  ttr.set_move(best_move);
  ttr.set_date(date);
  ttr.set_key(cp.get_hash());
  m_tt.write(ttr);
}

void Analysis::set_fertility(const Move& m, uint64_t nodes)
{
  if(m_ply==0){
    if(nodes>m_fertility[0].second){
      m_fertility[2]=m_fertility[1];
      m_fertility[1]=m_fertility[0];
      m_fertility[0].first=m,m_fertility[0].second=nodes;
    }else if(nodes>m_fertility[1].second){
      m_fertility[2]=m_fertility[1];
      m_fertility[1].first=m,m_fertility[1].second=nodes;      
    }else if(nodes>m_fertility[2].second){
      m_fertility[2].first=m,m_fertility[2].second=nodes;
    }
  }   
}

void Analysis::sort_fertility(MoveList& ml)
{
  for(size_t i=0;i<ml.size();i++){
    if(ml[i]==m_fertility[0].first){
      ml[i].xcode(fertility1);
    }else if(ml[i]==m_fertility[1].first){
      ml[i].xcode(fertility2);
    }else if(ml[i]==m_fertility[2].first){
      ml[i].xcode(fertility3);
    }
  }

  for(size_t i=0;i<ml.size();i++){
    for(size_t j=i;j<ml.size();j++){
      if(ml[j].xcode()<ml[i].xcode()){
	swap(ml[i],ml[j]);
      }
    }
  }
}
