/*

Copyright (C) 2018 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"threadpool.h"



ThreadPool::ThreadPool(int n,TranspositionTable& tt) : m_tt(tt)
{
  ChessPosition cp;
  m_number_of_threads=n;
  for(int i=0;i<n;i++){
    m_cp.push_back(cp);
    m_busy.push_back(false);
    m_pana.push_back(nullptr);
    m_pana.back()=new Analysis(m_cp.back(),m_tt,*this);
    m_pana.back()->set_thread_id(i);
    m_thread.push_back(thread(std::ref(*m_pana.back())));
    m_timer.push_back(Timer());
  }
}

ThreadPool::ThreadPool(const ThreadPool& tp) : m_tt(tp.m_tt)
{
  ChessPosition cp;
  m_number_of_threads=tp.m_number_of_threads;
  for(unsigned i=0;i<m_number_of_threads;i++){
    m_cp.push_back(cp);
    m_busy.push_back(false);
    m_pana.push_back(nullptr);
    m_pana.back()=new Analysis(m_cp.back(),m_tt,*this);
    m_pana.back()->set_thread_id(i);
    m_thread.push_back(thread(std::ref(*m_pana.back())));
    m_timer.push_back(Timer());
  }
}

ThreadPool::~ThreadPool()
{
  clear();
}

Analysis* ThreadPool::operator[](int i)const
{
  return m_pana[i];
}

void ThreadPool::clear()
{
  std::lock_guard<mutex> lock(m_mutex_threadpool);
  m_busy.clear();
  for(auto p : m_pana){p->quit();}
  for(auto& t : m_thread){t.join();}
  m_thread.clear();
  m_timer.clear();
  for(auto& p : m_pana){delete p;}
  m_pana.clear();
  m_cp.clear();
  m_running_threads=0;
  m_number_of_threads=0;
}

int ThreadPool::size()const
{
  return m_number_of_threads;
}

void ThreadPool::set_size(const int& n)
{
  clear();
  std::lock_guard<mutex> lock(m_mutex_threadpool);
  ChessPosition cp;
  m_number_of_threads=n;
  m_free_thread=n>0?true:false;
  for(int i=0;i<n;i++){
    m_cp.push_back(cp);
    m_busy.push_back(false);
    m_pana.push_back(nullptr);
    m_pana.back()=new Analysis(m_cp.back(),m_tt,*this);
    m_pana.back()->set_thread_id(i);
    m_thread.push_back(thread(std::ref(*m_pana.back())));
    m_timer.push_back(Timer());
  }
}

long long ThreadPool::get_nodes()const
{
  long long r{0};
  for(auto p : m_pana){
    r+=p->get_node_count();
  }
  return r;
}

void ThreadPool::clear_nodes()
{
  for(auto& p : m_pana){
    p->clear_node_count();
  }
}

bool ThreadPool::open_new_thread(int& n)
{
  if(!m_free_thread){n=-1;return false;}

  std::lock_guard<mutex> lock(m_mutex_threadpool);
  if( (m_running_threads>=m_number_of_threads) || (!m_busy[0] && n!=0) ){
    n=-1;
    return false;
  }
  
  //----------------------------------------------
  for(unsigned i=0;i<m_number_of_threads;i++){
    if(m_busy[i]==false){
      m_busy[i]=true;
      m_timer[i].start();
      m_pana[i]->set_status(spinning);
      m_running_threads++;
      if(m_running_threads==m_number_of_threads){m_free_thread=false;}
      n=i;
      return true;
    }
  }
  //----------------------------------------------
  
  throw logic_error("ThreadPool::open_new_thread     threadpool error");
}

string ThreadPool::state()const
{
  std::lock_guard<mutex> lock(m_mutex_threadpool);  
  ostringstream oss;
  oss << "available threads: " << m_number_of_threads << endl;
  oss << "running threads:   " << m_running_threads << endl;
  oss << "free thread:       " << boolalpha << m_free_thread << endl;
  oss << "spinning threads:  ";
  int spin{0};
  for(unsigned i=0;i<m_number_of_threads;i++){
    if(m_pana[i]->get_status()==spinning || m_pana[i]->get_status()==finished){spin++;}
  }
  oss << spin << endl;
  oss << "busy thread:  ";
  for(unsigned i=0;i<m_busy.size();i++){
    if(m_busy[i]){
      oss << "  " << i;
    }
  }
  oss << endl;
  for(unsigned i=0;i<m_timer.size();i++){
    long long nodes=m_pana[i]->get_thread_nodes();
    int used=m_pana[i]->get_thread_time().milliseconds();
    int total_i=m_timer[i].total_time().milliseconds();
    int percentage=total_i>1?100.0*used/total_i:0;
    int total=m_timer[0].total_time().milliseconds();
    int percentage2=total>1?total_i*100/total:0;
    float nps=used>1?nodes*1000.0/used:0;
    oss << "thread(" << i << ")" << "   nodes: " << nodes << "\t\tthinking time: " << used
	<< "\t\trunning time: " << total
	<< "\t\tthinking: " << percentage << "%\t\trunning: " << percentage2 << "%\t\tnps: " << nps << endl; 
  }
  return oss.str();
}

void ThreadPool::nw_search(const int& n, float depth, int alpha, int beta, int lmr, int reductions, int extensions, const Analysis& x)
{
  *(m_pana[n])=x;
  m_pana[n]->set_thread_depth(depth);
  m_pana[n]->set_thread_alpha(alpha);
  m_pana[n]->set_thread_beta(beta);
  m_pana[n]->set_thread_lmr(lmr);
  m_pana[n]->set_thread_reductions(reductions);
  m_pana[n]->set_thread_extensions(extensions);
  m_pana[n]->set_action(start_nws);
}

void ThreadPool::result(const int& i,int& value,int& nodes)
{
  if(i<0)return;
  while(m_pana[i]->get_status()!=finished){};
  std::lock_guard<mutex> lock(m_mutex_threadpool);  
  m_pana[i]->set_status(spinning);
  value=m_pana[i]->get_thread_score();
  nodes=m_pana[i]->get_node_count();
  m_running_threads--;
  m_free_thread=true;
  m_timer[i].stop();
  m_busy[i]=false;
}


bool ThreadPool::is_spinning(const int& i)const
{
  return m_pana[i]->get_status()==spinning;
}

void ThreadPool::stop_spinning()
{
  for(unsigned i=1;i<m_pana.size();i++){
    if(m_pana[i]->get_status()==finished || m_pana[i]->get_status()==spinning){m_pana[i]->set_status(idle);}
  }
}
